#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Test case for an ingest path with files like those in
/lustre/aoc/cluster/pipeline/dsoc-dev/workspaces/staging/tmpx_ratuqh"""

import json
import shutil
import tarfile
from pathlib import Path

import pytest
from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.manifest_components import (
    MANIFEST_FILENAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    IngestionManifestKey,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope

from .conftest import (
    ADDITIONAL_METADATA_FILENAME,
    ANCILLARY_PRODUCTS,
    ingest_path,
    populate_fake_tmpx_ratuqh_ingest_path,
)

# pylint: disable=E0401, E0402, W0621


# pylint: disable=R1721


def test_manifest_picks_up_ing_artifact(ingest_path: Path):
    """
    There's an ingestion artifacts tar in the ingest path,
    and it should show up among the ancillary products.

    :return:
    """

    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)

    _, manifest_file = build_tmpx_ratuqh_image_manifest(ingest_path)

    # make sure that tar really is there
    candidates = [file for file in ingest_path.glob("ingestion_artifacts*.tar")]
    assert len(candidates) > 0
    ing_arties = candidates[0]

    with open(manifest_file, "r") as infile:
        manifest_contents = infile.read()
        assert ing_arties.name in manifest_contents

    shutil.rmtree(ingest_path)


def test_addl_metadata_not_at_bottom_of_manifest(ingest_path: Path):
    """
    Manifest should have only parameters, input group, and output group
    :param ingest_path:
    :return:
    """
    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)
    manifest, _ = build_tmpx_ratuqh_image_manifest(ingest_path)
    mf_json = manifest.to_json()
    keys = ["parameters", "input_group", "output_group"]
    assert len(mf_json) == len(keys)
    for key in keys:
        assert key in mf_json.keys()


def test_manifest_picks_up_pip_artifact(ingest_path: Path):
    """
    There's an pipeline artifacts tar in the ingest path,
    and it should show up among the ancillary products.

    :return:
    """
    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)
    manifest, _ = build_tmpx_ratuqh_image_manifest(ingest_path)
    maybe_pips = [file for file in ingest_path.glob(AncillaryProductType.PIPELINE_ARTIFACTS.value + "*.tar")]
    pip_artie = maybe_pips[0]

    mf_json = manifest.to_json()
    og_json = mf_json[IngestionManifestKey.OUTPUT_GROUP.value]

    ap_jsons = og_json[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
    found = False
    for ap_json in ap_jsons:
        if ap_json["filename"] == pip_artie.name:
            found = True
            break

    assert found

    shutil.rmtree(ingest_path)


def test_catches_invalid_sp_type():
    """
    If we try to give IngestionManifestBuilder a science product type that doesn't translate into
    one of the ScienceProductType enum values, it should yell at us.

    :return:
    """

    additional_metadata = AbstractTextFile(filename="does-this-image-make-me-look-fat.json", content="")

    # we expect this to work fine
    IngestionManifestBuilder(
        staging_source_dir=Path(),
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="somewhere-over-the-rainbow",
        additional_metadata=additional_metadata,
    )
    # this, however, should fail
    with pytest.raises(ValueError):
        IngestionManifestBuilder(
            staging_source_dir=Path(),
            telescope=Telescope.EVLA.value,
            sp_type="model_built_out_of_playdoh",
            locator="somewhere-over-the-rainbow",
            additional_metadata=additional_metadata,
        )


def test_catches_invalid_telescope():
    """
    If we try to give IngestionManifestBuilder a telescope that doesn't translate into
    one of the Telescope enum values, have no mercy.

    :return:
    """

    additional_metadata = AbstractTextFile(filename="arcane-knowledge.json", content="")

    # we expect this to work fine
    IngestionManifestBuilder(
        staging_source_dir=Path(),
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="ill-have-what-shes-having",
        additional_metadata=additional_metadata,
    )
    # this, however, should go belly-up
    with pytest.raises(ValueError):
        IngestionManifestBuilder(
            staging_source_dir=Path(),
            telescope="nathans_mars_telescope",
            sp_type=ScienceProductType.IMAGE.value,
            locator="ill-have-what-shes-having",
            additional_metadata=additional_metadata,
        )


def test_ancillary_products_rendered_correctly(ingest_path: Path):
    """
    The ancillary_products section of the manifest we build
    should match the one in the example.

    :param ingest_path: staging source dir
    :return:
    """
    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)

    manifest, _ = build_tmpx_ratuqh_image_manifest(ingest_path)

    output_group = manifest.output_group
    aps = output_group.ancillary_products
    assert len(aps) == len(ANCILLARY_PRODUCTS)
    ap_data = []

    for ap in aps:
        ap_data.append({"filename": ap.filename, "json": ap.to_json()})
    assert len(ap_data) == len(aps)

    # make sure all the ancillary products were created...
    filename_count = 0
    for product in ANCILLARY_PRODUCTS:
        # (It won't find the ingest artifact tar yet, because we didn't populate
        # the ingest path with it; it's produced during manifest creation.)

        # ...and that each one's JSON is well formed.
        for properties in ap_data:
            if properties["filename"] == product.filename:
                filename_count += 1
                # If there's a class `repr` in there rather than JSON-serializable text,
                # dump will fail
                json.dumps(properties["json"])

    ingest_artifacts_tar = find_ingestion_artifacts_tar(ingest_path)
    assert ingest_artifacts_tar
    filename_count += 1

    assert filename_count == len(ANCILLARY_PRODUCTS)

    shutil.rmtree(ingest_path)


def test_output_science_products_rendered_correctly(ingest_path: Path):
    """
    The output_group section of the manifest we build
    should match the one in the example:
    * a "science_products" section containing one science product comprising "type", "filename",
    and two ancillary products
    * an "ancillary products" section comprising three ancillary products

    :param ingest_path:
    :return:
    """
    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)
    manifest, _ = build_tmpx_ratuqh_image_manifest(ingest_path)
    mf_json = manifest.to_json()
    og_json = mf_json[IngestionManifestKey.OUTPUT_GROUP.value]
    print(og_json)

    # there should be a science product...
    sp_json = og_json[IngestionManifestKey.SCIENCE_PRODUCTS.value][0]
    assert len(sp_json) == 3
    for key in (IngestionManifestKey.ANCILLARY_PRODUCTS.value, "type", "filename"):
        assert key in sp_json.keys()

    # and ancillary products belonging to the science product...
    sp_ap_jsons = sp_json[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
    assert len(sp_ap_jsons) == 2

    # ... and ancillary products twisting in the wind all by themselves
    ap_jsons = og_json[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
    assert len(ap_jsons) == 3

    # there should be a weblog, the ingestion artifacts tar, and the pipeline artifacts tar
    filenames_found = []
    for ap_json in ap_jsons:
        filename = ap_json["filename"]
        if filename.endswith(TARFILE_EXT):
            if filename.startswith(AncillaryProductType.INGESTION_ARTIFACTS.value) or filename.startswith(
                AncillaryProductType.PIPELINE_ARTIFACTS.value
            ):
                filenames_found.append(filename)
        elif filename == WEBLOG_FILENAME:
            filenames_found.append(filename)

    assert len(filenames_found) == 3

    shutil.rmtree(ingest_path)


def test_ing_artifacts_tar_has_only_what_it_should(ingest_path: Path):
    """
    The ingestion_artifacts tar should contain ONLY
    the ingestion manifest and the additional metadata file

    :return:
    """
    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)
    manifest, _ = build_tmpx_ratuqh_image_manifest(ingest_path)
    addl_md = manifest.parameters.additional_metadata
    ing_arties_tar = find_ingestion_artifacts_tar(ingest_path)
    with tarfile.open(ing_arties_tar, "r") as tar:
        members = tar.getmembers()
        assert len(members) == 2
        for member in members:
            assert member.name.endswith(MANIFEST_FILENAME) or member.name.endswith(addl_md.filename)

    shutil.rmtree(ingest_path)


# -----------------------------
#  U  T  I  L  I  T  I  E  S
# -----------------------------


def build_tmpx_ratuqh_image_manifest(staging_source_dir: Path):
    """
    Build an image ingestion manifest that "should" match image_manifest_tmpx_ratuqh.json

    :return: the ingestion manifest and the manifest file in which it lives
    """

    additional_metadata = AbstractTextFile(filename=ADDITIONAL_METADATA_FILENAME, content="")

    return IngestionManifestBuilder(
        staging_source_dir=staging_source_dir,
        additional_metadata=additional_metadata,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="uid://evla/calibration/ea93dae5-3495-47fa-887d-4be2852f5f14",
    ).build()


def find_ingestion_artifacts_tar(staging_source_dir: Path):
    """
    There should be an ingestion artifacts tar after manifest creation.

    :param staging_source_dir:
    :return:
    """
    ing_artifacts_tars = [
        file
        for file in staging_source_dir.iterdir()
        if file.name.startswith(AncillaryProductType.INGESTION_ARTIFACTS.value) and file.name.endswith(TARFILE_EXT)
    ]
    assert len(ing_artifacts_tars) == 1
    return ing_artifacts_tars[0]
