#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import json
import os
import subprocess
from subprocess import CompletedProcess
from typing import List
from unittest.mock import MagicMock, patch

import pytest
from _pytest.logging import LogCaptureFixture
from hypothesis import given
from hypothesis import strategies as st
from wf_inspector.inspector import (
    CLASS_AD_BATCH_ID,
    CLASS_AD_CLUSTER_ID,
    ClassAd,
    NoIdForWorkflowError,
    NoJobFoundError,
    NoSuchWorkflowRequestError,
    get_container_id_by_substr,
    get_dag_job_id,
    get_running_class_ads,
    get_wf_request_htcondor_id,
    is_dag_job,
    list_running_jobs,
    ssh_to_condor_job,
)

TEST_CONTAINER_ID = "testid123"
TEST_CONTAINER_ID_BYTES = TEST_CONTAINER_ID.encode("ascii")
TEST_JOB_IDS = [1, 2, 3]
TEST_DAG_ID = 3
TEST_BATCH_ID = 4
TEST_DAG_JOB_ID = 5


def test_get_container_id_by_substr_found():
    fake_completed_process = CompletedProcess(
        ["docker", "ps", "-q", "-f", "workflow"], returncode=0, stdout=TEST_CONTAINER_ID
    )
    with patch("wf_inspector.inspector.subprocess.run", return_value=fake_completed_process):
        test_container_id = get_container_id_by_substr("workflow")

    assert test_container_id == TEST_CONTAINER_ID


def test_get_container_id_by_substr_not_found():
    fake_completed_process = CompletedProcess(["docker", "ps", "-q", "-f", "workflow"], returncode=0, stdout="")
    with patch("wf_inspector.inspector.subprocess.run", return_value=fake_completed_process):
        test_container_id = get_container_id_by_substr("workflow")

    assert test_container_id is None


def test_get_container_id_by_substr_error(caplog):
    expected_logs = [
        "Docker command not accessible. Docker is either not installed or you lack permissions to use it.",
        "Are you running this from within a Docker container? Docker is not installed in containers.",
    ]
    fake_completed_process = CompletedProcess(["docker", "ps", "-q", "-f", "workflow"], returncode=1, stdout="")
    with patch("wf_inspector.inspector.subprocess.run", return_value=fake_completed_process):
        test_container_id = get_container_id_by_substr("workflow")

    assert test_container_id is None
    # Assert all expected logs have been logged out
    assert all(log in [record.message for record in caplog.records] for log in expected_logs)


def test_get_wf_request_htcondor_id_docker():
    # Environment save point
    capo_profile_savepoint = os.environ.get("CAPO_PROFILE")
    # Set environment to check that the URL is set correctly
    os.environ["CAPO_PROFILE"] = "docker"

    fake_response = MagicMock()
    fake_response.status_code = 200
    fake_response.json.return_value = {"htcondor_job_id": "-1"}

    with patch("wf_inspector.inspector.requests.get", return_value=fake_response) as mock_get:
        test_htcondor_id = get_wf_request_htcondor_id("-1")
        mock_get.assert_called_with("http://localhost:3456/workflows/name/requests/-1/htcondor_id")

    assert test_htcondor_id == -1

    # Return environment to previous state
    if capo_profile_savepoint is None:
        capo_profile_savepoint = ""
    os.environ["CAPO_PROFILE"] = capo_profile_savepoint


def test_get_wf_request_htcondor_id_nondocker():
    # Environment save point
    capo_profile_savepoint = os.environ.get("CAPO_PROFILE")
    # Set environment to check that the URL is set correctly
    os.environ["CAPO_PROFILE"] = "dsoc-dev"

    fake_response = MagicMock()
    fake_response.status_code = 200
    fake_response.json.return_value = {"htcondor_job_id": "-1"}

    fake_capo_config = MagicMock()
    fake_capo_config.settings.return_value.serviceUrl = "https://ws-dev.nrao.edu:3456"

    with patch("wf_inspector.inspector.requests.get", return_value=fake_response) as mock_get:
        with patch(
            "wf_inspector.inspector.CapoConfig",
            return_value=fake_capo_config,
        ):
            test_htcondor_id = get_wf_request_htcondor_id("-1")
            mock_get.assert_called_with("https://ws-dev.nrao.edu:3456/workflows/name/requests/-1/htcondor_id")

    assert test_htcondor_id == -1

    # Return environment to previous state
    if capo_profile_savepoint is None:
        capo_profile_savepoint = ""
    os.environ["CAPO_PROFILE"] = capo_profile_savepoint


def test_get_wf_request_htcondor_id_exc():
    # Environment save point
    capo_profile_savepoint = os.environ.get("CAPO_PROFILE")
    # Set environment to check that the URL is set correctly
    os.environ["CAPO_PROFILE"] = "dsoc-dev"

    # NoIdForWorkflowError
    fake_response = MagicMock()
    fake_response.status_code = 200
    fake_response.json.return_value = {"htcondor_job_id": "None"}

    fake_capo_config = MagicMock()
    fake_capo_config.settings.return_value.serviceUrl = "https://ws-dev.nrao.edu:3456"

    with patch("wf_inspector.inspector.requests.get", return_value=fake_response) as mock_get:
        with patch(
            "wf_inspector.inspector.CapoConfig",
            return_value=fake_capo_config,
        ):
            with pytest.raises(NoIdForWorkflowError):
                get_wf_request_htcondor_id("-1")
            mock_get.assert_called_with("https://ws-dev.nrao.edu:3456/workflows/name/requests/-1/htcondor_id")

    # NoSuchWorkflowRequestError
    fake_response = MagicMock()
    fake_response.status_code = 500
    fake_response.json.return_value = None

    fake_capo_config = MagicMock()
    fake_capo_config.settings.return_value.serviceUrl = "https://ws-dev.nrao.edu:3456"

    with patch("wf_inspector.inspector.requests.get", return_value=fake_response) as mock_get:
        with patch(
            "wf_inspector.inspector.CapoConfig",
            return_value=fake_capo_config,
        ):
            with pytest.raises(NoSuchWorkflowRequestError):
                get_wf_request_htcondor_id("-1")
            mock_get.assert_called_with("https://ws-dev.nrao.edu:3456/workflows/name/requests/-1/htcondor_id")

    # Return environment to previous state
    if capo_profile_savepoint is None:
        capo_profile_savepoint = ""
    os.environ["CAPO_PROFILE"] = capo_profile_savepoint


@given(
    st.lists(
        st.fixed_dictionaries(
            {
                CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                "DiskUsage": st.integers(),
                CLASS_AD_BATCH_ID: st.integers(),
            }
        ),
        unique_by=lambda entry: entry[CLASS_AD_CLUSTER_ID],
        min_size=1,
    )
)
def test_get_running_class_ads(fake_running_jobs_json: List[ClassAd]):
    fake_completed_process = CompletedProcess(
        ["docker", "exec", "-it", TEST_CONTAINER_ID_BYTES, "condor_q", "-json"],
        returncode=0,
        stdout=json.dumps(fake_running_jobs_json),
    )
    with patch("wf_inspector.inspector.subprocess.run", return_value=fake_completed_process):
        actual_running_jobs_json = get_running_class_ads(TEST_CONTAINER_ID_BYTES)

    assert actual_running_jobs_json == fake_running_jobs_json


@given(
    st.lists(
        st.fixed_dictionaries(
            {
                CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                "DiskUsage": st.integers(),
                CLASS_AD_BATCH_ID: st.integers(),
            }
        ),
        unique_by=lambda entry: entry[CLASS_AD_CLUSTER_ID],
        min_size=3,
        max_size=3,
    )
)
def test_is_dag_job(fake_running_jobs_json: List[ClassAd]):
    """
    Given an HTCondor job ID
    When the ClassAd JSON is requested for the ID
    And the "JobBatchId" field exists
    Then it is a DAG job
    """
    assert is_dag_job(fake_running_jobs_json, TEST_DAG_ID)


@given(
    st.lists(
        st.fixed_dictionaries(
            {
                CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                "DiskUsage": st.integers(),
            }
        ),
        unique_by=lambda entry: entry[CLASS_AD_CLUSTER_ID],
        min_size=3,
        max_size=3,
    )
)
def test_is_not_dag_job(fake_running_jobs_json: List[ClassAd]):
    """
    Given an HTCondor job ID
    When the ClassAd JSON is requested for the ID
    And the "JobBatchId" field does not exist
    Then it is not a DAG job
    """
    assert not is_dag_job(fake_running_jobs_json, TEST_DAG_ID)


@given(
    st.lists(
        st.fixed_dictionaries(
            {
                CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                "DiskUsage": st.integers(),
            }
        ),
        unique_by=lambda entry: entry[CLASS_AD_CLUSTER_ID],
        min_size=3,
        max_size=3,
    )
)
def test_is_dag_job_exc(fake_running_jobs_json: List[ClassAd]):
    """
    Given an HTCondor job ID
    When the ClassAd JSON is requested for the ID
    And the "JobBatchId" field does not exist
    Then it is not a DAG job
    """
    with pytest.raises(NoJobFoundError):
        is_dag_job(fake_running_jobs_json, -1)


@given(
    st.lists(
        st.one_of(
            st.fixed_dictionaries(
                {
                    CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                    "DiskUsage": st.integers(),
                    CLASS_AD_BATCH_ID: st.sampled_from(TEST_JOB_IDS),
                }
            ),
            st.fixed_dictionaries(
                {
                    CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                    "DiskUsage": st.integers(),
                }
            ),
        ),
        unique_by=lambda entry: entry[CLASS_AD_CLUSTER_ID],
        min_size=3,
        max_size=3,
    ),
    st.fixed_dictionaries(
        {
            CLASS_AD_CLUSTER_ID: st.just(TEST_BATCH_ID),
            "DiskUsage": st.integers(),
            CLASS_AD_BATCH_ID: st.just(TEST_BATCH_ID),
        }
    ),
    st.fixed_dictionaries(
        {
            CLASS_AD_CLUSTER_ID: st.just(TEST_DAG_JOB_ID),
            "DiskUsage": st.integers(),
            CLASS_AD_BATCH_ID: st.just(TEST_BATCH_ID),
        }
    ),
)
def test_get_dag_job_id(fake_running_jobs_json: List[ClassAd], dag_json: ClassAd, dag_job_json: ClassAd):
    """
    Given the ID of a running DAGMan job
    When the ClassAd JSON is requested for the cluster
    Then return the first JSON entry found that:
    * Contains the field "JobBatchId"
    * "JobBatchId" field is equal to the DAGMan job ID
    """
    all_json = fake_running_jobs_json + [dag_json, dag_job_json]
    assert get_dag_job_id(all_json, TEST_BATCH_ID) == TEST_DAG_JOB_ID


@given(
    st.lists(
        st.one_of(
            st.fixed_dictionaries(
                {
                    CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                    "DiskUsage": st.integers(),
                    CLASS_AD_BATCH_ID: st.sampled_from(TEST_JOB_IDS),
                }
            ),
            st.fixed_dictionaries(
                {
                    CLASS_AD_CLUSTER_ID: st.sampled_from(TEST_JOB_IDS),
                    "DiskUsage": st.integers(),
                }
            ),
        ),
        unique_by=lambda entry: entry[CLASS_AD_CLUSTER_ID],
        min_size=3,
        max_size=3,
    ),
    st.fixed_dictionaries(
        {
            CLASS_AD_CLUSTER_ID: st.just(TEST_BATCH_ID),
            "DiskUsage": st.integers(),
            CLASS_AD_BATCH_ID: st.just(TEST_BATCH_ID),
        }
    ),
    st.fixed_dictionaries(
        {
            CLASS_AD_CLUSTER_ID: st.just(TEST_DAG_JOB_ID),
            "DiskUsage": st.integers(),
            CLASS_AD_BATCH_ID: st.just(TEST_BATCH_ID),
        }
    ),
)
def test_get_dag_job_id_exc(fake_running_jobs_json: List[ClassAd], dag_json: ClassAd, dag_job_json: ClassAd):
    """
    Given the ID of a running DAGMan job
    When the ClassAd JSON is requested for the cluster
    Then return the first JSON entry found that:
    * Contains the field "JobBatchId"
    * "JobBatchId" field is equal to the DAGMan job ID
    """
    all_json = fake_running_jobs_json + [dag_json, dag_job_json]
    with pytest.raises(NoJobFoundError):
        get_dag_job_id(all_json, -1)


def test_ssh_to_condor_job():
    with patch("wf_inspector.inspector.subprocess.run") as mock_subprocess:
        ssh_to_condor_job(TEST_CONTAINER_ID_BYTES, -1)
        mock_subprocess.assert_called_with(
            ["docker", "exec", "-it", TEST_CONTAINER_ID_BYTES, "condor_ssh_to_job", "-auto-retry", "-1"], check=True
        )


def test_list_running_jobs(caplog: LogCaptureFixture):
    expected_logs = [
        "HTCondor job ID not found!",
        "Here are the jobs currently running on the cluster\n",
        "Is your job still running?",
    ]
    with patch("wf_inspector.inspector.subprocess.run") as mock_subprocess:
        list_running_jobs(TEST_CONTAINER_ID_BYTES)
        mock_subprocess.assert_called_with(
            ["docker", "exec", "-it", TEST_CONTAINER_ID_BYTES, "condor_q"], stdout=subprocess.PIPE, check=True
        )
    logged_messages = caplog.messages
    assert all(log in logged_messages for log in expected_logs)
